package com.inxar.jenesis;

/**
 * $Id: MCodeWriter.java,v 1.2 2006-02-16 21:59:50 wee Exp $
 *
 * Copyright (C) 2000, 2001  Paul Cody Johnston - pcj@inxar.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
import java.util.Vector;
import java.io.PrintWriter;
import org.inxar.jenesis.*;

/**
 * Standard <code>CodeWriter</code> implementation.
 */
public class MCodeWriter
	implements CodeWriter
{
	// the actual writer
	private PrintWriter out;
	private int colNo;
	private int lineNo;
	private int indentNo;
	private boolean isLineNew;
	private boolean hasQueue;
	private Vector queue;
    public MCodeWriter(PrintWriter out)
    {
	this.out = out;
	colNo = 0;
	lineNo = 1;		// start at line 1
	indentNo = 0;
	isLineNew = true;
	queue = new Vector();
    }
    /**
     * Decrements the tab and calls newLine()
     */
    public CodeWriter dedentLine()
    {
	if (hasQueue) {
	    flushQueue();
	    hasQueue = false;
	}
	indentNo--;
	newLine();
	return this;
    }
    private void flushQueue()
    {
	while(!queue.isEmpty()) {
	    Codeable c = (Codeable)queue.elementAt(0);
	    queue.removeElementAt(0);
	    write(c);
	}
    }
    /**
     * Returns the current number of characters in the
     * current line.  It does not take the indent into account.
     * Therefore, only the write methods and the space method
     * increment the column counter.
     */
    public int getColumnNumber()
    {
	return colNo;
    }
    /**
     * Returns the current number of indentation levels.
     */
    public int getIndentNumber()
    {
	return indentNo;
    }
    /**
     * Returns the number of lines of the current document.
     */
    public int getLineNumber()
    {
	return lineNo;
    }
    /**
     * Increments the tab and calls newLine()
     */
    public CodeWriter indentLine()
    {
	if (hasQueue) {
	    flushQueue();
	    hasQueue = false;
	}
	indentNo++;
	newLine();
	return this;
    }
    /**
     * Returns true if no characters have been written since the last
     * call of newLine().
     */
    public boolean isLineNew()
    {
	return isLineNew;
    }
    /**
     * Adds a the newLine string according to
     * <CODE>System.getProperty("line.separator")</CODE> and the
     * line is padded with the n tab characters where n is the number
     * returned by <code>getIndentNumber()</code>.
     */
    public CodeWriter newLine()
    {
	if (hasQueue) {
	    flushQueue();
	    hasQueue = false;
	}
	out.println();
	writeIndent();
	colNo = 0;
	lineNo++;
	isLineNew = true;
	return this;
    }
    /**
     * This method allows those codeable objects to inject a comment
     * without interrupting the line-by-line code itself.  For example, if
     * an expression wants to express a comment, it cannot do it until the
     * end of the line.  This method accepts a string argument.  Before the
     * newline is called, all comments given to the code writer will be
     * written after newLine has been called.
     */
    public void queue(Comment comment)
    {
	if (comment != null) {
	    comment.setText(comment.getText() + " >> see (ln,col) (" + lineNo + ", " + colNo + ")");
	    queue.addElement(comment);
	    hasQueue = true;
	}
    }
    /**
     * Resets the tab counter to zero and calls the newLine() method.
     */
    public CodeWriter resetLine()
    {
	indentNo = 0;
	newLine();
	return this;
    }
    /**
     * Writes a single space.
     */
    public CodeWriter space()
    {
	out.print(' ');
	colNo++;
	isLineNew = false;
	return this;
    }
    /**
     * Writes an array of characters.
     */
    public CodeWriter write(char[] chars)
    {

	if (chars != null) {
	    // print the chars;
	    out.print(chars);
	    // add
	    colNo += chars.length;
	    isLineNew = false;
	}
	return this;
    }
    /**
     * Writes an array of characters.
     */
    public CodeWriter write(char[] chars, int off, int len)
    {
	if (chars != null) {
	    // print the chars;
	    out.write(chars, off, len);
	    // add
	    colNo += chars.length;
	    isLineNew = false;
	}
	return this;
    }
    /**
     * Writes each element of the given object array.
     */
    public CodeWriter write(Object[] ao)
    {
	if (ao != null) {
	    for (int i=0; i<ao.length; i++)
		write(ao[i]);
	}
	return this;
    }
    /**
     * Iterates the array and sends each element to
     * <code>write(Codeable)</code>.
     */
    public CodeWriter write(Codeable[] aico)
    {
	if (aico != null)
	    for (int i=0; i<aico.length; i++)
		write(aico[i]);
	return this;
    }
    /**
     * Writes a single character.
     */
    public CodeWriter write(char c)
    {
	// print the char
	out.print(c);
	// add one
	colNo++;
	isLineNew = false;
	return this;
    }
    /**
     * Writes a double.
     */
    public CodeWriter write(double d)
    {
	out.print(d);
	colNo += Double.toString(d).length();
	isLineNew = false;
	return this;
    }
    /**
     * Writes a float.
     */
    public CodeWriter write(float f)
    {
	out.print(f);
	colNo += Float.toString(f).length();
	isLineNew = false;
	return this;
    }
    /**
     * Writes an integer.
     */
    public CodeWriter write(int i)
    {
	out.print(i);
	colNo += Integer.toString(i).length();
	isLineNew = false;
	return this;
    }
    /**
     * Writes an object.
     */
    public CodeWriter write(Object o)
    {
	if (o != null) {
	    if (o instanceof Codeable) {
		((Codeable)o).toCode(this);
	    } else {
		out.print(o);
		colNo += o.toString().length();
		isLineNew = false;
	    }
	}
	return this;
    }
    /**
     * Writes a string.
     */
    public CodeWriter write(String s)
    {
	if (s != null) {
	    out.print(s);
	    colNo += s.length();
	    isLineNew = false;
	}
	
	return this;
    }
    /**
     * Writes each object element of the given enumeration.
     */
    public CodeWriter write(java.util.Enumeration aEnum)
    {
	if (aEnum != null)
	    while(aEnum.hasMoreElements())
		write(aEnum.nextElement());
	return this;
    }
    /**
     * Instead of calling the <code>Object.toString()</code> method,
     * the <code>Object.toCode(public CodeWriter)</code> method is invoked with
     * <code>this</code> as the argument.
     */
    public CodeWriter write(Codeable ico)
    {
	if (ico != null)
	    ico.toCode(this);
	return this;
    }
    /**
     * Writes each object element of the given aEnumeration.
     */
    public CodeWriter write(Iterator aEnum)
    {
	if (aEnum != null)
	    while(aEnum.hasNext())
		write(aEnum.next());
	return this;
    }
    /**
     * Writes a boolean.
     */
    public CodeWriter write(boolean b)
    {
	// print it
	out.print(b);
	// add if 4:'true' or 5:'false'
	colNo += b ? 4 : 5;
	isLineNew = false;
	return this;
    }
    private void writeIndent()
    {
	int n = indentNo;
	while(n-->0)
	    out.print("    ");
    }
}
