package com.inxar.jenesis;

/**
 * $Id: MExpression.java,v 1.1 2003-08-25 12:31:54 pra Exp $
 *
 * Copyright (C) 2000, 2001  Paul Cody Johnston - pcj@inxar.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
import java.util.Vector;
import org.inxar.jenesis.*;
import com.inxar.jenesis.util.EnumerationIterator;


/**
 * Standard <code>Expression</code> implementations.
 */
public abstract class MExpression
	extends MVM.MCodeable
	implements Expression
{
	// ===============================================================
	// BLANK EXPRESSION
	// ===============================================================
	static class MBlank
	extends MExpression
	implements Blank
	{
	MBlank(MVM vm)
	{
	    super(vm);
	}

	public Type getType()
	{
	    return null;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    return out;
	}
	}

	// ===============================================================
	// CAST EXPRESSION
	// ===============================================================
	static class MCast
	extends MExpression
	implements Cast
	{
	MCast(MVM vm, Type type, Expression val)
	{
	    super(vm);
	    this.type = type;
	    this.val = val;
	}

	public void setExpression(Expression val)
	{
	    this.val = val;
	}

	public Expression getExpression()
	{
	    return val;
	}

	public Type getType()
	{
	    return type;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);
	    if (type==null) {
	    	out.write('(').write(val).write(')'); 
	    }else {
	    	out.write('(').write('(').write(type).write(')').write(val).write(')'); 
	    }
	    return out;
	}

	Type type;
	Expression val;
	}

	// ===============================================================
	// UNARY EXPRESSION
	// ===============================================================
	static class MUnary
	extends MExpression
	implements Unary
	{
	MUnary(MVM vm, int type)
	{
	    super(vm);
	    this.type = type;
	}

	MUnary(MVM vm, int type, Expression val)
	{
	    super(vm);
	    this.type = type;
	    this.val = val;
	}

	public void setValue(Expression val)
	{
	    this.val = val;
	}

	public Expression getValue()
	{
	    return val;
	}

	public int type()
	{
	    return type;
	}

	public Type getType()
	{
	    return val.getType();
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);

	    switch (type) {
	    case Unary.GROUP: out.write('(').write(val).write(')'); break;
	    case Unary.NOT: out.write('!').write(val); break;
	    case Unary.BITWISE_NOT: out.write('~').write(val); break;
	    case Unary.POS: out.write('+').write(val); break;
	    case Unary.NEG: out.write('-').write(val); break;
	    case Unary.AI: out.write("++").write(val); break;
	    case Unary.PI: out.write(val).write("++"); break;
	    case Unary.AD: out.write("--").write(val); break;
	    case Unary.PD: out.write(val).write("--"); break;
	    default: throw new RuntimeException("Please use a constant from the Unary interface when making new unary expressions.");
	    }

	    return out;
	}

	int type;
	Expression val;
	}

	// ===============================================================
	// BINARY EXPRESSION
	// ===============================================================
	static class MBinary
	extends MExpression
	implements Binary
	{
	MBinary(MVM vm, int type)
	{
	    super(vm);
	    this.type = type;
	}

	MBinary(MVM vm, int type, Expression l, Expression r)
	{
	    super(vm);
	    this.type = type;
	    this.l = l;
	    this.r = r;
	}

	public int type()
	{
	    return type;
	}

	public Type getType()
	{
	    return l.getType();
	}

	public void setLValue(Expression l)
	{
	    this.l = l;
	}

	public Expression getLValue()
	{
	    return l;
	}

	public void setRValue(Expression r)
	{
	    this.r = r;
	}

	public Expression getRValue()
	{
	    return r;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);

	    switch (type) {
	    case Binary.LAND: out.write(l).write(" && ").write(r); break;
	    case Binary.LOR: out.write(l).write(" || ").write(r); break;
	    case Binary.BAND: out.write(l).write(" & ").write(r); break;
	    case Binary.BOR: out.write(l).write(" | ").write(r); break;
	    case Binary.XOR: out.write(l).write(" ^ ").write(r); break;
	    case Binary.LEFT: out.write(l).write(" << ").write(r); break;
	    case Binary.RIGHT: out.write(l).write(" >> ").write(r); break;
	    case Binary.UNSIGNED: out.write(l).write(" >>> ").write(r); break;
	    case Binary.ADD: out.write(l).write(" + ").write(r); break;
	    case Binary.SUB: out.write(l).write(" - ").write(r); break;
	    case Binary.MUL: out.write(l).write(" * ").write(r); break;
	    case Binary.DIV: out.write(l).write(" / ").write(r); break;
	    case Binary.MOD: out.write(l).write(" % ").write(r); break;
	    case Binary.EQ: out.write(l).write(" == ").write(r); break;
	    case Binary.NE: out.write(l).write(" != ").write(r); break;
	    case Binary.GT: out.write(l).write(" > ").write(r); break;
	    case Binary.GTE: out.write(l).write(" >= ").write(r); break;
	    case Binary.LT: out.write(l).write(" < ").write(r); break;
	    case Binary.LTE: out.write(l).write(" <= ").write(r); break;
	    case Binary.IOF: out.write(l).write(" instanceof ").write(r); break;
	    case Binary.CAT: out.write(l).write(" + ").write(r); break;
	    default: throw new RuntimeException("Please use a constant from the Binary interface when making new binary expressions.");
	    }
	    return out;
	}

	int type;
	Expression l;
	Expression r;
	}

	// ===============================================================
	// TERNARY EXPRESSION
	// ===============================================================
	static class MTernary
	extends MExpression
	implements Ternary
	{
	MTernary(MVM vm, int type)
	{
	    super(vm);
	    this.type = type;
	}

	MTernary(MVM vm, int type, Expression one, Expression two, Expression three)
	{
	    super(vm);
	    this.type = type;
	    this.one = one;
	    this.two = two;
	    this.three = three;
	}

	public int type()
	{
	    return type;
	}

	public Type getType()
	{
	    return two.getType();
	}

	public void setValue1(Expression one)
	{
	    this.one = one;
	}

	public Expression getValue1()
	{
	    return one;
	}

	public void setValue2(Expression two)
	{
	    this.two = two;
	}

	public Expression getValue2()
	{
	    return two;
	}

	public void setValue3(Expression three)
	{
	    this.three = three;
	}

	public Expression getValue3()
	{
	    return three;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);
	    out.write(one).write(" ? ").write(two).write(" : ").write(three);
	    return out;
	}

	int type;
	Expression one;
	Expression two;
	Expression three;
	}

	// ===============================================================
	// ASSIGNMENT EXPRESSION
	// ===============================================================
	static class MAssign
	extends MBinary
	implements Assign
	{
	MAssign(MVM vm, int type)
	{
	    super(vm, type);
	}

	MAssign(MVM vm, int type, Variable l, Expression r)
	{
	    super(vm, type, l, r);
	}

	public int type()
	{
	    return type;
	}

	public void setVariable(Variable v)
	{
	    super.setLValue(v);
	}

	public Variable getVariable()
	{
	    return (Variable)super.getLValue();
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    out.queue(comment);

	    switch (type) {
	    case Assign.S: out.write(l).write(" = ").write(r); break;
	    case Assign.BAND: out.write(l).write(" &= ").write(r); break;
	    case Assign.BOR: out.write(l).write(" |= ").write(r); break;
	    case Assign.XOR: out.write(l).write(" ^= ").write(r); break;
	    case Assign.LEFT: out.write(l).write(" <<= ").write(r); break;
	    case Assign.RIGHT: out.write(l).write(" >>= ").write(r); break;
	    case Assign.UNSIGNED: out.write(l).write(" >>>= ").write(r); break;
	    case Assign.ADD: out.write(l).write(" += ").write(r); break;
	    case Assign.SUB: out.write(l).write(" -= ").write(r); break;
	    case Assign.MUL: out.write(l).write(" *= ").write(r); break;
	    case Assign.DIV: out.write(l).write(" /= ").write(r); break;
	    case Assign.MOD: out.write(l).write(" %= ").write(r); break;
	    case Assign.CAT: out.write(l).write(" += ").write(r); break;
	    default: throw new RuntimeException("Please use a constant from the Assign interface when making new Assignment expressions.");
	    }
	    return out;
	}
	}

	// ===============================================================
	// VARIABLE EXPRESSION
	// ===============================================================
	static class MVariable
  	extends MExpression
  	implements Variable
	{
	MVariable(MVM vm)
	{
	    super(vm);
	}

	MVariable(MVM vm, String name)
	{
	    super(vm);
	    this.name = name;
	}

	public void setName(String name)
	{
	    this.name = name;
	}

	public String getName()
	{
	    return name;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);
	    out.write(name);
	    return out;
	}

	public Type getType()
	{
	    return null;
	}

	String name;
	}

	// ===============================================================
	// ARRAY INITIALIZATION EXPRESSION
	// ===============================================================
	static class MArrayInitializer
	extends MExpression
	implements ArrayInitializer
	{
	MArrayInitializer(MVM vm, Object o)
	{
	    super(vm);
	    this.o = o;
	}

	public void setArgs(Object o)
	{
	    this.o = o;
	}

	public Object getArgs()
	{
	    return o;
	}

	public Type getType()
	{
	    return null;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);

	    // normal case
	    if (o == null) {
		out.write("{}");
	    } else if (o instanceof Object[]) {
		Object[] ao = (Object[])o;
		if (ao.length == 0) {
		    out.write("{}");
		} else {
		    writeElements(ao, out);
		}

		// the exceptional case where the primary object
		// is NOT an array, and we simply write the sole
		// element
	    } else if (o instanceof Expression) {
		out.write("{ ").write(o).write(" }");
	    }

	    return out;
	}

	void writeElement(Object o, CodeWriter out, boolean isNotFirst)
	{
	    if (o instanceof Object[]) {
		if (isNotFirst) out.write(", ");
		writeElements((Object[])o, out);
	    } else if (o instanceof Expression) {
		if (isNotFirst) out.write(", ");
		out.write(o);
	    }
	}

	void writeElements(Object[] ao, CodeWriter out)
	{
	    out.write('{');

	    // write each element, sending true if this
	    // is not the first element we are writing
	    for (int i=0; i<ao.length; i++)
		writeElement(ao[i], out, i>0);

	    out.write('}');
	}

	Object o;
	}

	// ===============================================================
	// ACCESSOR EXPRESSION
	// ===============================================================
	static class MAccessor
  	extends MExpression
  	implements Accessor
	{
	MAccessor(MVM vm)
	{
	    super(vm);
	}

	public MAccessor(MVM vm, String qual, String name)
	{
	    super(vm);
	    this.qual = qual;
	    this.name = name;
	}

	public void setName(String name)
	{
	    this.name = name;
	}

	public String getName()
	{
	    return name;
	}

	public void setQualifier(String qual)
	{
	    this.qual = qual;
	}

	public String getQualifier()
	{
	    return qual;
	}

	public Type getType()
	{
	    return null;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);
	    if (qual != null)
		out.write(qual).write('.');
	    out.write(name);
	    return out;
	}

	String name;
	String qual;
	}

	// ===============================================================
	// FIELD ACCESS EXPRESSION
	// ===============================================================
	static class MFieldAccess
  	extends MAccessor
  	implements FieldAccess
	{
	MFieldAccess(MVM vm)
	{
	    super(vm);
	}

	MFieldAccess(MVM vm, String qual, String name)
	{
	    super(vm, qual, name);
	}
	}

	// ===============================================================
	// ARRAY ACCESS EXPRESSION
	// ===============================================================
	static class MArrayAccess
  	extends MAccessor
  	implements ArrayAccess
	{
	MArrayAccess(MVM vm)
	{
	    super(vm);
	    dimensions = new Vector();
	}

	MArrayAccess(MVM vm, String qual, String name)
	{
	    super(vm, qual, name);
	    dimensions = new Vector();
	}

	public ArrayAccess addDim(Expression expr)
	{
	    dimensions.addElement(expr);
	    return this;
	}

	public Iterator getDims()
	{
	    return new EnumerationIterator(dimensions.elements(), dimensions.size());
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);

			for (int i=0; i<dimensions.size(); i++) {
				out.write('[').write(dimensions.elementAt(i)).write(']');
			}

	    return out;
	}

	Vector dimensions;
	}

	// ===============================================================
	// NARY EXPRESSION
	// ===============================================================
	abstract static class MNary
	extends MAccessor
	{
	MNary(MVM vm)
	{
	    super(vm);
	    ve = new Vector();
	}

	public Iterator getArgs()
	{
	    return new EnumerationIterator(ve.elements(), ve.size());
	}

	Vector ve;
	}

	// ===============================================================
	// METHOD INVOCATION EXPRESSION
	// ===============================================================
	static class MInvoke
  	extends MNary
  	implements Invoke
	{
	MInvoke(MVM vm, String qual, String name)
	{
	    super(vm);
	    this.qual = qual;
	    this.name = name;
	}

	public Invoke addArg(Expression arg)
	{
	    ve.addElement(arg);
	    return this;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);

	    out.write('(');
	    for (int i=0; i<ve.size(); i++) {
		if (i>0) out.write(", ");
		out.write(ve.elementAt(i));
	    }
	    out.write(')');

	    return out;
	}
	}

	// ===============================================================
	// CLASS INSTANCE CREATION EXPRESSION
	// ===============================================================
	static class MNewClass
  	extends MNary
  	implements NewClass
	{
	MNewClass(MVM vm, Type type)
	{
	    super(vm);
	    this.type = type;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);

	    out.write("new ").write(type).write('(');
	    for (int i=0; i<ve.size(); i++) {
		if (i>0) out.write(", ");
		out.write(ve.elementAt(i));
	    }
	    out.write(')');

	    return out;
	}

	public NewClass addArg(Expression arg)
	{
	    ve.addElement(arg);
	    return this;
	}

	public Type getType()
	{
	    return type;
	}

	public void setType(Type type)
	{
	    this.type = type;
	}

	Type type;
	}

	// ===============================================================
	// ARRAY INSTANCE CREATION EXPRESSION
	// ===============================================================
	static class MNewArray
  	extends MNary
  	implements NewArray
	{
	MNewArray(MVM vm, Type type)
	{
	    super(vm);
	    this.type = type;
	    dimensions = new Vector(); 	// dimensions
	}

	public Type getType()
	{
	    return type;
	}

	public void setType(Type type)
	{
	    this.type = type;
	}

	public Iterator getDims()
	{
	    // convert the vector contents to an array
	    return new EnumerationIterator(dimensions.elements(), dimensions.size());
	}

	public NewArray addDim(Expression e)
	{
	    dimensions.addElement(e);
	    return this;
	}

	public NewArray addDim()
	{
	    dimensions.addElement(new MBlank(vm));
	    return this;
	}

	public ArrayInitializer getInitializer()
	{
	    return ai;
	}

	public void setInitializer(ArrayInitializer ai)
	{
	    this.ai = ai;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    out.write("new ").write(type);

	    for (int i=0; i<dimensions.size(); i++) {
		out.write('[').write(dimensions.elementAt(i)).write(']');
	    }
	    out.write(ai);
	    return out;
	}

	Type type;
	Vector dimensions;
	ArrayInitializer ai;
	}

	// ===============================================================
	// ANONYMOUS CLASS CREATION EXPRESSION
	// ===============================================================
	static class MNewAnonymousClass
  	extends MDeclaration.MClassDeclaration
  	implements NewAnonymousClass
	{
	MNewAnonymousClass(MVM vm, Type type)
	{
	    super(vm, null);
	    ve = new Vector();
	    this.type = type;
	}

	public Iterator getArgs()
	{
	    return new EnumerationIterator(ve.elements(), ve.size());
	}

	public NewClass addArg(Expression arg)
	{
	    ve.addElement(arg);
	    return this;
	}

	public Type getType()
	{
	    return type;
	}

	public void setType(Type type)
	{
	    this.type = type;
	}

	public void setQualifier(String qual)
	{
	    this.qual = qual;
	}

	public String getQualifier()
	{
	    return qual;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    if (comment != null)
		out.queue(comment);

	    out.write("new ").write(type).write('(');

	    for (int i=0; i<ve.size(); i++) {
		if (i>0) out.write(", ");
		out.write(ve.elementAt(i));
	    }
	    out.write(')');

	    writeBlock(out, vm.getStyle("anonymous-class"));

	    return out;
	}

	Type type;
	Vector ve;
	String qual;
	}

	// ===============================================================
	// FREEFORM EXPRESSION
	// ===============================================================
	static class MFreeform
	extends MExpression
	implements Freeform
	{
	MFreeform(MVM vm)
	{
	    super(vm);
	    init();
	}

	MFreeform(MVM vm, String code)
	{
	    this(vm);
	    write(code);
	}

	void init()
	{
	    this.buf = new java.io.StringWriter();
	    this.out = new java.io.PrintWriter(buf);
	}

	public void setCode(String code)
	{
	    init();
	    buf.write(code);
	}

	public String getCode()
	{
	    return buf.toString();
	}

	public Type getType()
	{
	    return null;
	}

	public CodeWriter toCode(CodeWriter out)
	{
	    super.toCode(out);
	    out.write(buf.toString());
	    return out;
	}

	public Freeform write(boolean b)
	{
	    // print it
	    out.print(b);
	    // add if 4:'true' or 5:'false'
	    colNo += b ? 4 : 5;
	    isLineNew = false;
	    return this;
	}

	public Freeform write(char c)
	{
	    // print the char
	    out.print(c);
	    // add one
	    colNo++;
	    isLineNew = false;
	    return this;
	}

	public Freeform write(char[] chars)
	{

	    if (chars != null)
		{
				// print the chars;
		    out.print(chars);
				// add
		    colNo += chars.length;
		    isLineNew = false;
		}
	    return this;
	}

	public Freeform write(int i)
	{
	    out.print(i);
	    colNo += Integer.toString(i).length();
	    isLineNew = false;
	    return this;
	}

	public Freeform write(float f)
	{
	    out.print(f);
	    colNo += Float.toString(f).length();
	    isLineNew = false;
	    return this;
	}


	public Freeform write(double d)
	{
	    out.print(d);
	    colNo += Double.toString(d).length();
	    isLineNew = false;
	    return this;
	}

	public Freeform write(String s)
	{
	    if (s != null)
		{
		    out.print(s);
		    colNo += s.length();
		    isLineNew = false;
		}

	    return this;
	}

	public Freeform write(Object o)
	{
	    if (o != null)
		{
		    out.print(o);
		    colNo += o.toString().length();
		    isLineNew = false;
		}
	    return this;
	}

	public Freeform write(Object[] ao)
	{
	    if (ao != null) {
		for (int i=0; i<ao.length; i++)
		    write(ao[i]);
	    }
	    return this;
	}

	public Freeform space()
	{
	    out.print(' ');
	    colNo++;
	    isLineNew = false;
	    return this;
	}

	public Freeform newLine()
	{
	    out.println();
	    writeIndent();
	    colNo = 0;
	    lineNo++;
	    isLineNew = true;
	    return this;
	}

	public Freeform resetLine()
	{
	    indentNo = 0;
	    newLine();
	    return this;
	}

	public Freeform indentLine()
	{
	    indentNo++;
	    newLine();
	    return this;
	}

	public Freeform dedentLine()
	{
	    indentNo--;
	    newLine();
	    return this;
	}

	public boolean isLineNew()
	{
	    return isLineNew;
	}

	public int getIndentNumber()
	{
	    return indentNo;
	}

	public int getLineNumber()
	{
	    return lineNo;
	}

	public int getColumnNumber()
	{
	    return colNo;
	}

	private void writeIndent()
	{
	    int n = indentNo;
	    while(n-->0)
		out.print('\t');
	}

	// the actual writer
	java.io.StringWriter buf;
	java.io.PrintWriter out;
	int colNo;
	int lineNo;
	int indentNo;
	boolean isLineNew;
	}
    MExpression(MVM vm)
    {
	super(vm);
    }
    public CodeWriter toCode(CodeWriter out)
    {
	out.queue(comment);
	return out;
    }
}
