package com.inxar.jenesis;

/**
 * $Id: MLiteral.java,v 1.1 2003-08-25 12:31:54 pra Exp $
 *
 * Copyright (C) 2000, 2001  Paul Cody Johnston - pcj@inxar.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
import java.util.Vector;
import org.inxar.jenesis.*;
import com.inxar.jenesis.util.*;

/**
 * Standard <code>Literal</code> implementations.
 */
public abstract class MLiteral
	extends MVM.MCodeable
	implements Literal
{
	Comment comment;
	Type type;
	Object val;
	String label;

	// ===============================================================
	// NULL LITERAL
	// ===============================================================
	static class MNull
	extends MLiteral
	implements Null
	{
	MNull(MVM vm)
	{
	    super(vm, vm.NULL, null, "null");
	}
	}

	// ===============================================================
	// BOOLEAN LITERAL
	// ===============================================================
	static class MBooleanLiteral
	extends MLiteral
	implements BooleanLiteral
	{
	MBooleanLiteral(MVM vm, boolean val)
	{
	    super(vm, vm.BOOLEAN, new Boolean(val), (val ? "true" : "false"));
	}

	public boolean toBoolean() { return ((Boolean)val).booleanValue(); }
	}

	// ===============================================================
	// TRUE LITERAL
	// ===============================================================
	static class MTrue
	extends MBooleanLiteral
	implements True
	{
	MTrue(MVM vm)
	{
	    super(vm, true);
	}
	}

	// ===============================================================
	// FALSE LITERAL
	// ===============================================================
	static class MFalse
	extends MBooleanLiteral
	implements False
	{
	MFalse(MVM vm)
	{
	    super(vm, false);
	}
	}

	// ===============================================================
	// BYTE LITERAL
	// ===============================================================
	static class MByteLiteral
	extends MLiteral
	implements ByteLiteral
	{
	MByteLiteral(MVM vm, byte val)
	{
	    super(vm, vm.BYTE, new Byte(val), "(byte)"+val);
	}
	public byte toByte() { return ((Byte)val).byteValue(); }
	}

	// ===============================================================
	// SHORT LITERAL
	// ===============================================================
	static class MShortLiteral
	extends MLiteral
	implements ShortLiteral
	{
	MShortLiteral(MVM vm, short val)
	{
	    super(vm, vm.SHORT, new Short(val), "(short)"+val);
	}
	public short toShort() { return ((Short)val).shortValue(); }
	}

	// ===============================================================
	// INT LITERAL
	// ===============================================================
	static class MIntLiteral
	extends MLiteral
	implements IntLiteral
	{
	MIntLiteral(MVM vm, int val)
	{
	    super(vm, vm.INT, new Integer(val), Integer.toString(val));
	}
	public int toInt() { return ((Integer)val).intValue(); }
	}

	// ===============================================================
	// LONG LITERAL
	// ===============================================================
	static class MLongLiteral
	extends MLiteral
	implements LongLiteral
	{
	MLongLiteral(MVM vm, long val)
	{
	    super(vm, vm.LONG, new Long(val), Long.toString(val)+'L');
	}
	public long toLong() { return ((Long)val).longValue(); }
	}

	// ===============================================================
	// FLOAT LITERAL
	// ===============================================================
	static class MFloatLiteral
	extends MLiteral
	implements FloatLiteral
	{
	MFloatLiteral(MVM vm, float val)
	{
	    super(vm, vm.FLOAT, new Float(val), Float.toString(val)+"F");
	}
	public float toFloat() { return ((Float)val).floatValue(); }
	}

	// ===============================================================
	// DOUBLE LITERAL
	// ===============================================================
	static class MDoubleLiteral
	extends MLiteral
	implements DoubleLiteral
	{
	MDoubleLiteral(MVM vm, double val)
	{
	    super(vm, vm.DOUBLE, new Double(val), Double.toString(val)+"D");
	}
	public double toDouble() { return ((Double)val).doubleValue(); }
	}

	// ===============================================================
	// CHAR LITERAL
	// ===============================================================
	static class MCharLiteral
	extends MLiteral
	implements CharLiteral
	{
	MCharLiteral(MVM vm, char val)
	{
	    super(vm, vm.CHAR, new Character(val), "'"+val+"'");
	}
	public char toChar() { return ((Character)val).charValue(); }
	}

	// ===============================================================
	// OCTAL LITERAL
	// ===============================================================
	static class MOctalLiteral
	extends MCharLiteral
	implements OctalLiteral
	{
	MOctalLiteral(MVM vm, char val)
	{
	    super(vm, val);
	    this.label = "'\\"+Integer.toOctalString(val)+"'";
	}
	}

	// ===============================================================
	// UNICODE LITERAL
	// ===============================================================
	static class MUnicodeLiteral
	extends MCharLiteral
	implements UnicodeLiteral
	{
	MUnicodeLiteral(MVM vm, char val)
	{
	    super(vm, val);
	    this.label = "'\\u"+Integer.toHexString(val)+"'";
	}
	}

	// ===============================================================
	// SCIENTIFIC LITERAL
	// ===============================================================
	static class MScientificLiteral
	extends MLiteral
	implements ScientificLiteral
	{
	MScientificLiteral(MVM vm, int precision, int scale, int exponent)
	{
	    super(vm, vm.DOUBLE, null, ""+precision+"."+scale+"e"+exponent);
	    this.precision = precision;
	    this.scale = scale;
	    this.exponent = exponent;
	}

	public int getPrecision() { return precision; }
	public int getScale() { return scale; }
	public int getExponent() { return exponent; }
	int precision;
	int scale;
	int exponent;
	}

	// ===============================================================
	// STRING LITERAL
	// ===============================================================
	static class MStringLiteral
	extends MLiteral
	implements StringLiteral
	{
	MStringLiteral(MVM vm, String val)
	{
	    super(vm, vm.STRING, val, "\""+escape(val)+"\"");
	}
	public String toString() { return (String)val; }

	private static String escape(String s)
	{
	    //System.out.println("escaping now!");
	    // Cache the length of the String to espae.
	    int len = s.length();

	    // Make a new buffer at with 110% capacity such a single
	    // escape won't cause the stringbuffer to needlessly
	    // double.
	    StringBuffer b = new StringBuffer(len + len/10);

	    char c;
	    // Iterate over the length of the string.
	    for (int i = 0; i < len; i++) {

		// Get each char.
		c = s.charAt(i);

		// Check if the char is in the ACSII range.
		if (c >= 0x20  && c < 0x7f) {
	
		    // The char is ASCII, but we still may need to
		    // escape it.
		    switch (c) {
		    case '"' : b.append("\\\""); break;
		    case '\'': b.append("\\'" ); break;
		    case '\\': b.append("\\\\"); break;
		    default: b.append(c);
		    }

		} else {

		    // Need special escapes for these chars.
		    switch (c) {
		    case '\b': b.append("\\b"); break;
		    case '\t': b.append("\\t"); break;
		    case '\n': b.append("\\n"); break;
		    case '\f': b.append("\\f"); break;
		    case '\r': b.append("\\r"); break;
		    default: {
			//System.out.println("writing unicode escape for '"+c+"'");
			// Write it as a unicode escape.
			b.append("\\u");
			String n = Integer.toHexString((int)c);
			for (int j = n.length(); j < 4; j++) {
			    b.append('0');
			}
			b.append(n);
		    }
		    }
		}
	    }

	    // If no escapes or unicode chars we encountered, there's
	    // no need to mint a new String.
	    return b.length() > len ? b.toString() : s;
	}
	}

	// ===============================================================
	// CLASS LITERAL
	// ===============================================================
	static class MClassLiteral
	extends MLiteral
	implements ClassLiteral
	{
	MClassLiteral(MVM vm, ClassType val)
	{
	    super(vm, val, val, val.toString()+".class");
	}
	public Class toClass() { return null; }
	}
    MLiteral(MVM vm, Type type, Object val, String label)
    {
	super(vm);
	this.type = type;
	this.val = val;
	this.label = label;
    }
    public Comment getComment()
    {
	return comment;
    }
    public Type getType()
    {
	return type;
    }
    public void setComment(Comment comment)
    {
	this.comment = comment;
    }
    public CodeWriter toCode(CodeWriter out)
    {
	out.queue(comment);
	out.write(label);
	return out;
    }
    public Object toObject()
    {
	return val;
    }
}
