package com.inxar.jenesis.util;

/**
 * $Id: MemberComparator.java,v 1.1 2003-08-25 12:31:55 pra Exp $
 *
 * Copyright (C) 2000, 2001  Paul Cody Johnston - pcj@inxar.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
import org.inxar.jenesis.*;

/**
 * Comparator which is used to sort <code>Members</code> in a class.
 */
public class MemberComparator
	implements java.util.Comparator
{
	protected static final int CONSTANT = 3;
	protected static final int INTERFACE = 4;
	protected static final int CONSTRUCTOR = 5;
	protected static final int METHOD = 6;
	protected static final int FIELD = 7;
	protected static final int CLASS = 8;
	protected static final int STATIC_INIT = 9;
	protected static final int OTHER = 10;

	protected static final int EQUAL = 0;
	protected static final int LESS = -1;
	protected static final int GREATER = 1;

    public MemberComparator()
    {
    }
    protected int assignInt(Object o)
    {
	if      (o instanceof Interface)         return INTERFACE;
	else if (o instanceof Constructor)       return CONSTRUCTOR;
	else if (o instanceof Method)            return METHOD;
	else if (o instanceof Field) {
	    // Special case.  See if it's static and final (a
	    // constant).
	    Field f = (Field)o;
	    return f.isStatic() && f.isFinal() ? CONSTANT : FIELD;
	}
	else if (o instanceof ClassDeclaration)  return CLASS;
	else if (o instanceof StaticInitializer) return STATIC_INIT;
	else                                     return OTHER;
    }
    public int compare(Object o1, Object o2)
    {
	// assign an integer for the compared types
	int i1 = assignInt(o1);
	int i2 = assignInt(o2);

	// are they equal?...
	if (i1 == i2) {

	    // if so, we will fork to the appropriate type
	    // identity comparison method
	    switch (i1)
		{
		case CLASS: return compareClasses((ClassDeclaration)o1, (ClassDeclaration)o2);
		case METHOD: return compareMethods((Method)o1, (Method)o2);
		case FIELD: return compareFields((Field)o1, (Field)o2);
		case CONSTRUCTOR: return compareConstructors((Constructor)o1, (Constructor)o2);
		case INTERFACE: return compareInterfaces((Interface)o1, (Interface)o2);
		default: return EQUAL;
		}

	    // the types are not the same,
	    // and therefore we can compare directly
	} else {
	    return (i1 < i2) ? LESS : GREATER;
	}
    }
    /**
     * utility method to compare boolean state.  If they are different,
     * the first boolean arg (o1) is deemed GREATER if it is the true one.
     */
    protected int compareBooleans(boolean b1, boolean b2)
    {
	// if they are different, return the comparison, otw equal
	if (b1 ^ b2)
	    return b1 ? GREATER : LESS;
	else
	    return EQUAL;
    }
    protected int compareClasses(ClassDeclaration o1, ClassDeclaration o2)
    {
	int x = EQUAL;

	// compare access levels
	x = compareInts(o1.getAccess(), o2.getAccess());
	if (x != EQUAL) return x;

	// compare abstract state
	x = compareBooleans(o1.isAbstract(), o2.isAbstract());
	if (x != EQUAL) return x;

	// compare static state
	x = compareBooleans(o1.isStatic(), o2.isStatic());
	if (x != EQUAL) return x;

	// compare final state
	x = compareBooleans(o1.isFinal(), o2.isFinal());
	if (x != EQUAL) return x;

	// compare name state
	x = compareStrings(o1.getName(), o2.getName());
	if (x != EQUAL) return x;

	// pretty damn close enough
	return x;
    }
    protected int compareConstructors(Constructor o1, Constructor o2)
    {
	int x = EQUAL;

	// compare access levels
	x = compareInts(o1.getAccess(), o2.getAccess());
	if (x != EQUAL) return x;

	// ctors have same acces, same # of ctors. We could continue
	// and compare the arguments, but naw...
	return x;
    }
    protected int compareFields(Field o1, Field o2)
    {
	int x = EQUAL;

	// compare static state
	x = compareBooleans(o1.isStatic(), o2.isStatic());
	if (x != EQUAL) return x;

	// compare final state
	x = compareBooleans(o1.isFinal(), o2.isFinal());
	if (x != EQUAL) return x;

	// compare access levels
	x = compareInts(o1.getAccess(), o2.getAccess());
	if (x != EQUAL) return x;

	// compare name state
	x = compareStrings(o1.getName(), o2.getName());
	if (x != EQUAL) return x;

	// pretty damn close enough
	return x;
    }
    protected int compareInterfaces(Interface o1, Interface o2)
    {
	int x = EQUAL;

	// compare name state
	x = compareStrings(o1.getName(), o2.getName());
	if (x != EQUAL) return x;

	// pretty damn close enough
	return x;
    }
    /**
     * utility method to compare integer state.  If they are different,
     * the they are compared with GREATER deemed to the larger number.
     */
    protected int compareInts(int i1, int i2)
    {
	if (i1 == i2) return EQUAL;
	else return (i1 < i2) ? LESS : GREATER;
    }
    protected int compareMethods(Method o1, Method o2)
    {
	int x = EQUAL;

	// compare access levels
	x = compareInts(o1.getAccess(), o2.getAccess());
	if (x != EQUAL) return x;

	// compare static state
	x = compareBooleans(o1.isStatic(), o2.isStatic());
	if (x != EQUAL) return x;

	// compare final state
	x = compareBooleans(o1.isFinal(), o2.isFinal());
	if (x != EQUAL) return x;

	// compare name state
	x = compareStrings(o1.getName(), o2.getName());
	if (x != EQUAL) return x;

	// pretty damn close enough
	return x;
    }
    /**
     * utility method to compare strings. uses jdk1.2's compareTo
     * method.  'does s1 compared to s2', not the other way around.
     */
    protected int compareStrings(String s1, String s2)
    {
	// compare strings
	return s1.compareTo(s2);
    }
}
