package org.inxar.jenesis;

/**
 * $Id: VirtualMachine.java,v 1.3 2004-07-27 15:47:41 hob Exp $
 *
 * Copyright (C) 2000, 2001  Paul Cody Johnston - pcj@inxar.org
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */
import java.io.*;
import java.util.Properties;

/**
 * The <code>VirtualMachine</code> acts as a factory method for
 * <code>CompilationUnit</code>, <code>Type</code>,
 * <code>Literal</code>, and <code>Expression</code> objects.  Most
 * <code>Statement</code> factory methods are located on
 * <code>Block</code>; most <code>Declaration</code> factory methods
 * are located on the objects that contain them.  
 */
public abstract class VirtualMachine
{
	/**
	 * The default <code>VirtualMachine</code> implementation.
	 */
	protected static VirtualMachine virtualMachine;
	
	/**
	 * Name of the Charset used for Output encoding. 
	 * See http://www.iana.org/assignments/character-sets for possible values. 
	 */
	public static String charsetName = "ISO_8859-1";

    /**
     * Writes out the source code to all enclosed
     * <code>CompilationUnit</code> instances.  
     */
    public abstract void encode() throws java.io.IOException;
    /**
     * Returns the active compiler. 
     */
    public abstract VirtualCompiler getCompiler();
	/**
	 * Returns the default <code>VirtualMachine</code> implementation
	 * previously set using the static
	 * <code>setVirtualMachine(VirtualMachine)</code> method OR
	 * returns the reference implementation
	 * (<code>com.inxar.jenesis.MVM</code>.  
	 */
	public static VirtualMachine getVirtualMachine()
	{
	if (virtualMachine == null) {

	    InputStream in = null;

	    try {

		String filename = System.getProperty("jenesis.blockstyle-properties-filename");
		if (filename != null) {
		    in = new FileInputStream(filename);
		} else {
		    filename = "blockstyle.properties";
		    in = org.inxar.jenesis.VirtualMachine.class.getClassLoader().getResourceAsStream(filename);
		}
		
		if (in == null) 
		    throw new RuntimeException
			("Cannot instantiate VirtualMachine: " + 
			 "Could not find blockstyle.properties file in system classpath " + 
			 "or from -Djenesis.blockstyle-properties-filename!");
	    
		Properties p = new Properties();
		p.load(new BufferedInputStream(in));
		in.close();
		in = null;

		virtualMachine = new com.inxar.jenesis.MVM(p);

	    } catch (IOException ioex) {
		RuntimeException rex = new RuntimeException
		    ("Could not load VirtualMachine blockstyles: "+ioex.getMessage());
		ioex.printStackTrace();
		throw rex;
	    } finally {
		if (in != null) 
		    try { in.close(); in = null;} catch (Exception ex) {}
	    }
	}

	return virtualMachine;
	}
    /**
     * <code>Expression</code> factory method which returns a new
     * field access with the given qualifier and name.  
     */
    public abstract FieldAccess newAccess(String qualifier, String name);
    /**
     * <code>Expression</code> factory method which returns a new
     * anonymous class instance creation expression for the given
     * type.  
     */
    public abstract NewAnonymousClass newAnon(Type type);
    
    public abstract Annotation newAnnotation(String text);
    public abstract Annotation newAnnotation(String name, String text);
    
    /**
     * <code>Type</code> factory method which returns an array of
     * primitive type with the given constant and dimensions.  
     */
    public abstract ArrayType newArray(int type, int dims);
    /**
     * <code>Type</code> factory method which returns an array of
     * class type with the given name and dimensions.  
     */
    public abstract ArrayType newArray(String type, int dims);
    /**
     * <code>Expression</code> factory method which returns a new
     * array instance creation expression for the given type.  
     */
    public abstract NewArray newArray(Type type);
    /**
     * <code>Expression</code> factory method which returns a new
     * array access expression with the given qualifier and name.  
     */
    public abstract ArrayAccess newArrayAccess(String qual, String name);
    /**
     * <code>Expression</code> factory method which returns a new
     * array initializer with the given arguments.  
     */
    public abstract ArrayInitializer newArrayInit(Object o);
    /**
     * <code>Expression</code> factory method which returns a new
     * assignment of the given type.  
     */
    public abstract Assign newAssign(int type, Variable left, Expression right);
    /**
     * <code>Expression</code> factory method which returns a new
     * assignment of the SIMPLE type.  This is a convenience * method.  
     */
    public abstract Assign newAssign(Variable left, Expression right);
    /**
     * <code>Expression</code> factory method which returns a new
     * unary function of the given type.  
     */
    public abstract Binary newBinary(int type, Expression left, Expression right);
    /**
     * <code>Expression</code> factory method which returns the blank
     * expression.  
     */
    public abstract Blank newBlank();
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>byte</code> literal.  
     */
    public abstract ByteLiteral newByte(byte val);
    /**
     * <code>Expression</code> factory method which returns a new
     * cast function having the given type and value.  
     */
    public abstract Cast newCast(Type type, Expression value);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>char</code> literal.  
     */
    public abstract CharLiteral newChar(char val);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>Class</code> literal.  
     */
    public abstract ClassLiteral newClass(String className);
    /**
     * <code>Expression</code> factory method which returns a new
     * class instance creation expression for the given type.  
     */
    public abstract NewClass newClass(Type type);
    /**
     * Creates a new <code>CompilationUnit</code> in this
     * <code>Machine</code> and returns it.  
     */
    public abstract CompilationUnit newCompilationUnit(String sourcepath);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>double</code> literal.  
     */
    public abstract DoubleLiteral newDouble(double val);
    /**
     * <code>Literal</code> factory method which returns the boolean
     * <code>false</code> literal.  
     */
    public abstract False newFalse();
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>float</code> literal.  
     */
    public abstract FloatLiteral newFloat(float val);
    /**
     * <code>Expression</code> factory method which returns a new
     * freeform expression with the given code.  
     */
    public abstract Freeform newFree(String code);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>int</code> literal.  
     */
    public abstract IntLiteral newInt(int val);
    /**
     * <code>Expression</code> factory method which returns a new
     * method invocation with the given qualifier and name.  
     */
    public abstract Invoke newInvoke(String qualifier, String name);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>long</code> literal.  
     */
    public abstract LongLiteral newLong(long val);
    /**
     * <code>Literal</code> factory method which returns the
     * <code>null</code> literal.  
     */
    public abstract Null newNull();
    /**
     * <code>Literal</code> factory method which returns a new octal
     * <code>char</code> literal.  
     */
    public abstract OctalLiteral newOctal(char val);
    /**
     * <code>Literal</code> factory method which returns a new
     * scientific floating point <code>char</code> literal.  
     */
    public abstract ScientificLiteral newScientific(int precision, int scale, int exponent);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>short</code> literal.  
     */
    public abstract ShortLiteral newShort(short val);
    /**
     * <code>Literal</code> factory method which returns a new
     * <code>String</code> literal.  
     */
    public abstract StringLiteral newString(String val);
    /**
     * <code>Expression</code> factory method which returns a new
     * unary function of the given type.  
     */
    public abstract Ternary newTernary(int type, Expression one, Expression two, Expression three);
    /**
     * <code>Literal</code> factory method which returns the boolean
     * <code>true</code> literal.  
     */
    public abstract True newTrue();
    /**
     * <code>Type</code> factory method which returns a primitive type
     * named by the given <code>int</code> as allowed by the constants
     * listed in the <code>Type</code> interface.  
     */
    public abstract PrimitiveType newType(int type);
    /**
     * <code>Type</code> factory method which returns a class type
     * named with the given <code>String</code>.  
     */
    public abstract ClassType newType(String name);
    /**
     * <code>Expression</code> factory method which returns a new
     * unary function of the given type.  
     */
    public abstract Unary newUnary(int type, Expression value);
    /**
     * <code>Literal</code> factory method which returns a new unicode
     * <code>char</code> literal.  
     */
    public abstract UnicodeLiteral newUnicode(char val);
    /**
     * <code>Expression</code> factory method which returns a new
     * variable with the given name.  
     */
    public abstract Variable newVar(String name);
    /**
     * Sets the active compiler. 
     */
    public abstract void setCompiler(VirtualCompiler compiler);
    /**
     * Sets the default <code>VirtualMachine</code> implementation.
     */
    public static void setVirtualMachine(VirtualMachine vm)
    {
	virtualMachine = vm;
    }
}
